@@ -5,6 +5,7 @@ The imports are taking a substantial amount of time so loaders.py is
 lazily loaded.
 """
 
+import signal
 import sys
 import os
 import time
@@ -66,6 +67,7 @@ litellm = lazy_import.lazy_module("litellm")
 deepgram = lazy_import.lazy_module("deepgram")
 pydub = lazy_import.lazy_module("pydub")
 ffmpeg = lazy_import.lazy_module("ffmpeg")
+torchaudio = lazy_import.lazy_module("torchaudio")
 
 
 try:
@@ -148,6 +150,21 @@ if "pdftotext" in globals():
                 return "\n\n".join(pdftotext.PDF(f))
     pdf_loaders["pdftotext"] = pdftotext_loader_class
 
+# unsilence audio
+sox_effects = [
+    # ["norm"],  # normalize audio
+    # isolate voice frequency
+    # ["highpass", "-1", "100"],
+    # ["lowpass", "-1", "3000"],
+    # -2 is for a steeper filtering: removes high frequency and very low ones
+    # ["highpass", "-2", "50"],
+    # ["lowpass", "-2", "5000"],
+    ["norm"],  # normalize audio
+    # max silence should be 3s
+    ["silence", "-l", "1", "0", "1%", "-1", "3.0", "1%"],
+    ["norm"],
+]
+
 global_temp_dir = [None]  # will be replaced when load_one_doc is called
 
 
@@ -844,14 +861,7 @@ def load_logseq_markdown(debug: bool, path: str, file_hash: str) -> List[Documen
                 meta[k] = v
                 cont = cont.replace(f"{k}:: {v}", "").strip()
             cont = dedent(cont)
-            # note: should it be dedented? that saves a lot of token
-            # use tabs instead of spaces to save tokens and avoid confusion the LLM
-            lines = cont.splitlines()
-            lines = [li.expandtabs(tabsize=2) for li in lines if li.strip()]
-            lines = [li.replace("  ", "\t") for li in lines]
-            cont = "\n".join(lines)
-            if cont.count("\t") * 2 ==  cont.count("\t\t"):
-                cont = cont.replace("\t\t", "\t")
+            cont = cont.replace("\t", "    ")
             content += "\n" + cont
 
         doc = Document(
@@ -868,6 +878,7 @@ def load_local_audio(
     path: str,
     file_hash: str,
     audio_backend: str,
+    audio_unsilence: Optional[bool] = None,
 
     whisper_lang: Optional[str] = None,
     whisper_prompt: Optional[str] = None,
@@ -876,6 +887,46 @@ def load_local_audio(
     ) -> List[Document]:
     assert Path(path).exists(), f"file not found: '{path}'"
 
+    if audio_unsilence:
+        red(f"Removing silence from audio file {path.name}")
+        waveform, sample_rate = torchaudio.load(path)
+
+        dur = waveform.size[1] / sample_rate
+        start = time.time()
+        waveform, sample_rate = torchaudio.sox_effects.apply_effects_tensor(
+            waveform,
+            sample_rate,
+            sox_effects,
+            )
+        elapsed = time.time() - start
+        new_dur = waveform.size[1] / sample_rate
+        assert new_dur < dur, (
+            f"Failed to remove silence for {path.name}:\n"
+            f"Original duration: {dur:.1f}\n"
+            f"New duration: {new_dur:.1f}\n"
+        )
+        assert new_dur > 10, (
+            f"Silence removal ended up with a suspiciously short audio for {path.name}:\n"
+            f"Original duration: {dur:.1f}\n"
+            f"New duration: {new_dur:.1f}\n"
+        )
+        red(f"Removed silence from {path.name}: {dur:.1f} -> {new_dur:.1f} in {elapsed:.1f}s")
+
+        unsilenced_path = global_temp_dir[0] / f"unsilenced_audio_{uuid.uuid4()}.ogg"
+        assert not unsilenced_path.exists()
+        torchaudio.save(
+            uri=str(unsilenced_path.resolve().absolute()),
+            sample_rate=sample_rate,
+            src=waveform,
+            format="ogg",
+        )
+        unsilenced_hash=file_hasher({"path": unsilenced_path})
+
+        old_path = path
+        old_hash = file_hash
+        path = unsilenced_path
+        file_hash = unsilenced_hash
+
     if audio_backend == "whisper":
         assert deepgram_kwargs is None, "Found kwargs for deepgram but selected whisper backend for local_audio"
         content = transcribe_audio_whisper(
@@ -933,6 +984,7 @@ def load_local_video(
     path: str,
     file_hash: str,
     audio_backend: str,
+    audio_unsilence: Optional[bool] = None,
 
     whisper_lang: Optional[str] = None,
     whisper_prompt: Optional[str] = None,
@@ -942,15 +994,17 @@ def load_local_video(
     assert Path(path).exists(), f"file not found: '{path}'"
 
     audio_path = global_temp_dir[0] / f"audio_from_video_{uuid.uuid4()}.mp3"
-    assert not Path(audio_path).exists()
+    assert not audio_path.exists()
 
     # extract audio from video
     try:
         whi(f"Exporting audio from {path} to {audio_path} (this can take some time)")
         t = time.time()
-        stream = ffmpeg.input(path)
-        stream = ffmpeg.output(stream, audio_path)
-        ffmpeg.run(stream)
+        ffmpeg.input(
+            path
+        ).output(
+            str(audio_path.resolve().absolute())
+        ).run()
         whi(f"Done extracting audio in {time.time()-t:.2f}s")
     except Exception as err:
         red(f"Error when getting audio from video using ffmpeg. Retrying with pydub. Error: '{err}'")
@@ -979,6 +1033,7 @@ def load_local_video(
         whisper_lang=whisper_lang,
         whisper_prompt=whisper_prompt,
         deepgram_kwargs=deepgram_kwargs,
+        audio_unsilence=audio_unsilence,
     )
 
 
@@ -1353,6 +1408,11 @@ def load_pdf(
     # probability
     probs = {}
 
+    def timeout_handler(signum, frame):
+        "crash if a pdf loader is taking too much time"
+        raise TimeoutError
+    pdf_timeout = 5 * 60  # in seconds
+
     pbar = tqdm(total=len(pdf_loaders), desc=f"Parsing PDF {name}", unit="loader")
     for loader_name in pdf_loaders:
         pbar.desc = f"Parsing PDF {name} with {loader_name}"
@@ -1360,7 +1420,11 @@ def load_pdf(
             if debug:
                 red(f"Trying to parse {path} using {loader_name}")
 
+            signal.signal(signal.SIGALRM, timeout_handler)
+            signal.alarm(pdf_timeout)
             content = _pdf_loader(loader_name, path, file_hash)
+            signal.alarm(0)
+
             pbar.update(1)
 
             if "unstructured" in loader_name.lower():
@@ -1397,6 +1461,7 @@ def load_pdf(
                 # time on running all the others
                 break
         except Exception as err:
+            signal.alarm(0)
             yel(f"Error when parsing '{path}' with {loader_name}: {err}")
             if "content" not in locals():
                 pbar.update(1)
